// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut fr Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file ElementFunction.h */

#ifndef AMDIS_ELEMENTFUNCTION_H
#define AMDIS_ELEMENTFUNCTION_H

#include "AbstractFunction.h"
#include "DOFVector.h"
#include "BasisFunction.h"

namespace AMDiS {

  /// Abstract access to functions living on elements.
  template<typename T>
  class ElementFunction : public AbstractFunction<T, DimVec<double> >
  {
  public:
    /// constructor.
    ElementFunction() 
      : elInfo(NULL) 
    {}

    /// destructor.
    virtual ~ElementFunction() {}

    /// sets \ref elInfo_;
    inline void setElInfo(const ElInfo *elInfo_) 
    { 
      elInfo = elInfo_; 
    }

  protected:
    /// ElInfo the function currently lives on.
    const ElInfo *elInfo;
  };

  /// ElementFunction wich encapsulates the evaluation of an analytical function.
  template<typename T>
  class ElementFunctionAnalytic : public ElementFunction<T>
  {
  public:
    /// constructor
    ElementFunctionAnalytic(const AbstractFunction<T, WorldVector<double> > *fct_)
      : ElementFunction<T>(),
	fct(fct_)
    {}

    /// evaluation at given coordinates.
    T operator()(const DimVec<double>& bary) const 
    {
      WorldVector<double> worldCoords;
      this->elInfo->coordToWorld(bary, worldCoords);
      return (*fct)(worldCoords);
    }

  protected:
    /// function to be avaluated at world coordinates.
    const AbstractFunction<T, WorldVector<double> > *fct;
  };


  /// ElementFunction wich encapsulates the interpolation of an DOFVector.
  template<typename T>
  class ElementFunctionDOFVec : public ElementFunction<T>
  {
  public:
    /// constructor.
    ElementFunctionDOFVec(const DOFVector<T> *vec) 
      : ElementFunction<T>(),
	dofVector(vec)
    {}

    /// evaluation at given coordinates.
    T operator()(const DimVec<double>& bary) const 
    {
      T* localVec = new T[dofVector->getFESpace()->getBasisFcts()->getNumber()];
      dofVector->getLocalVector(this->elInfo->getElement(), localVec);
      T t = dofVector->getFESpace()->getBasisFcts()->evalUh(bary, localVec);

      delete [] localVec;
      return t;
    }

  protected:
    /// DOFVector to be interpolated.
    const DOFVector<T> *dofVector;
  };

}

#endif
