// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut fr Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file PollutionError.h */

/** \defgroup Estimator Estimator module
 * @{ <img src="estimator.png"> @}
 */

#ifndef AMDIS_POLLUTIONERROR_KONVEX_H
#define AMDIS_POLLUTIONERROR_KONVEX_H

#include <list>
#include <time.h>
#include "ResidualEstimator.h"
#include "PartitionElementData.h"

namespace AMDiS {

  /** \ingroup Estimator
   * \brief Estimator for parallel solutions. 
   *
   * estimates the error using following formulae
   * \f$||e_h||_{1,\Omega_0}=C_1 ||u_h||_{H_1(\Omega_1)}+
   * \frac{C_2}{d^2}||u_h||_{L_2(\Omega\setminus\Omega_1}\f$
   * where \f$\Omega\f$ is the complete area of interest and \f$\Omega_1\f$ the area of 
   * this process  with overlapping. \f$\Omega_0\$ is the area of this process without 
   * overlapping,\f$d\$ contains the distance 
   * dist(\f$\partial \Omega_0,\partial \Omega_1\f$)
   */
  class PollutionErrorKonvex : public ResidualEstimator 
  {
  public:
    class Creator : public EstimatorCreator 
    {
    public:
      Creator() : EstimatorCreator() {}

      virtual ~Creator() {}

      /// Returns a new ODirSolver object.
      Estimator* create() 
      { 
	return new PollutionErrorKonvex(name, row);
      }
    };
    
    PollutionErrorKonvex(std::string name, int r);

    //sets the overlapping width for 2d-elements
    virtual void init(double timestep);

    virtual void estimateElement(ElInfo *elInfo);

    virtual void exit(bool output);

  protected:
    /// constant in front of the inner residual   
    double C1p;
    
    /// constant in front of the outer residual    
    double C2p;

    ///traverse-class for edges and there normals in 2d
    class BoundaryTraverseStack: public std::list< WorldVector<double>* >
    {
    public:
      BoundaryTraverseStack(Mesh *mesh, PartitionStatus p);			
      
    private:
      PartitionStatus ps;
      bool proofConditions(ElInfo *elInfo);
    };

    /// computes the squared distance of element el to the boundary of \f$\Omega_0$\f$
    double getDistance(ElInfo* el);

    /// stack to traverse over the boundary of \f$\Omega_0$\f$
    BoundaryTraverseStack* bts;

    /// barycentric center
    DimVec<double> center;
  };
}
#endif
